#ifndef _IPT_SNAPSHOT_H
#define _IPT_SNAPSHOT_H

#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/spinlock.h>

//#define __DEBUG_SNAPSHOT__
#ifdef __DEBUG_SNAPSHOT__
#define DEBUGP(fmt, args...) printk(fmt, ##args)
#else
#define DEBUGP(fmt, args...) {}
#endif

/*
 * should be synchronized with userspace snapshot.h !!!
 */

#define IPPROTO_SNAPSHOT             240
#define IPOPT_SNAPSHOT               (100 | IPOPT_CONTROL)

// snapshot message type (1..255)
#define SNAPSHOT_MSGTYPE_RESET        1
#define SNAPSHOT_MSGTYPE_CLEAR        2
#define SNAPSHOT_MSGTYPE_CONTROL      3
#define SNAPSHOT_MSGTYPE_VMAP         4
#define SNAPSHOT_MSGTYPE_RMAP         5
#define SNAPSHOT_MSGTYPE_MARKER       6
#define SNAPSHOT_MSGTYPE_SNAPSHOT     7

#define SNAPSHOT_MSGTYPE_CONTROL_ACK  (SNAPSHOT_MSGTYPE_CONTROL | 0x80)
#define SNAPSHOT_MSGTYPE_VMAP_ACK     (SNAPSHOT_MSGTYPE_VMAP    | 0x80)
#define SNAPSHOT_MSGTYPE_RMAP_ACK     (SNAPSHOT_MSGTYPE_RMAP    | 0x80)

#define SNAPSHOT_MSGTYPE_SMAP         10
#define SNAPSHOT_MSGTYPE_STAT         11
#define SNAPSHOT_MSGTYPE_SMAP_ACK     (SNAPSHOT_MSGTYPE_SMAP    | 0x80)

// snapshot state
#define SNAPSHOT_STATE_UNINITIALIZED  0
#define SNAPSHOT_STATE_READY          1
#define SNAPSHOT_STATE_STARTED        2
#define SNAPSHOT_STATE_FINISHED       3

#define MAX_LINK_LIMIT                50
#define MAX_STAT_LIMIT                100

static const char* snapshot_state_desc[] = {
  "UNINITIALIZED", "READY", "STARTED", "FINISHED"
};

/*
static const char* snapshot_msgtype_desc[] = {
  "NULL", "RESET", "CLEAR", "CONTROL", "VMAP", "RMAP", "MARKER", "SNAPSHOT" 
};
*/

#define SNAPSHOT_OPTLEN (sizeof(struct snapshot_opt))
#define SNAPSHOT_IPHLEN (sizeof(struct iphdr) + SNAPSHOT_OPTLEN)
#define SNAPSHOT_OPTPTR(iph) \
  ((struct snapshot_opt *) (((__u8 *) iph) + sizeof(struct iphdr)))
#define SNAPSHOT_DATAPTR(iph) (((__u8 *) iph) + SNAPSHOT_IPHLEN)

#define for_each_link(recp, i) \
  for (i=0; i<(recp)->linkcnt; i++)

#define for_each_stat(recp, i) \
  for (i=0; i<(recp)->statcnt; i++)

// snapshot IP option
struct snapshot_opt {
  __u8 type;            // IPOPT_SNAPSHOT
  __u8 length;          // SNAPSHOT_OPTLEN
  __u8 msgtype;         // SNAPSHOT_MSGTYPE_XXX
  __u8 padding;         // zero-padding
  __u32 seqno;
};

// snapshot message header of VMAP
struct snapshot_msghdr_vmap {
  __u16 entrycnt;
  __u16 snapshotcnt;
};

// snapshot message header of RMAP
struct snapshot_msghdr_rmap {
  __u16 entrycnt;
  __u8 connected;       // 1: direct-connected to victim, 0: otherwise
  __u8 padding;
};

// snapshot message header of SMAP
struct snapshot_msghdr_smap {
  __u16 entrycnt;
  __u16 padding;
};

// snapshot message header of SNAPSHOT
struct snapshot_msghdr_snapshot {
  __u16 entrycnt;
  __u16 padding;
};

// snapshot message header of STAT
struct snapshot_msghdr_stat {
  __u16 entrycnt;
  __u16 padding;
};

// snapshot state
struct snapshot_state {
  __u8 state;
  __u32 cnt;

  __u32 addr;
  unsigned char dev_addrlen;
  unsigned char dev_addr[MAX_ADDR_LEN];
};

// snapshot record
struct snapshot_rec {
  spinlock_t lock;

  __u32 seqno;
  __u32 victim_addr;
  __u16 snapshotcnt;
  __u32 packetcnt;
  int finished;

  struct snapshot_state local;

  int linkcnt;
  int maxlinkcnt;
  struct snapshot_state *link;

  int maxstatcnt;
  int statcnt;
  struct snapshot_statistic *stat;
};

struct snapshot_statistic {
  __u32 addr;
  __u32 cnt;
};

static inline void DEBUGP_DEV_ADDR(unsigned char *dev_addr, int dev_addrlen)
{
  int j;
  for (j=0; j<dev_addrlen; j++) {
    DEBUGP("%2.2x%c", dev_addr[j], (j==dev_addrlen-1) ? '\n' : ':');
  }
}

#define DEBUGP_SSMSG(prefix, desc, in, iph, opt) {                      \
    DEBUGP("%s: recv %s (%u) from %u.%u.%u.%u/",                        \
           prefix, desc, ntohl(opt->seqno), NIPQUAD(iph->saddr));       \
    DEBUGP_DEV_ADDR((unsigned char *) in->dev_addr, in->addr_len); }

#define ssrec_lock(ssrec, flags) spin_lock_irqsave(&(ssrec.lock), flags)
#define ssrec_unlock(ssrec, flags) spin_unlock_irqrestore(&(ssrec.lock), flags)

inline __u8 snapshot_msgtype(struct iphdr *iph);
int snapshot_send_marker(struct snapshot_rec *rec, __u32 saddr, __u32 daddr);
int snapshot_send_snapshot(struct snapshot_rec *rec, __u32 saddr, __u32 daddr);
int snapshot_send_stat(struct snapshot_rec *rec, __u32 saddr, __u32 daddr);
int snapshot_send_ack(struct snapshot_rec *rec, __u32 saddr, __u32 daddr, int msgtype);
int snapshot_parse_vmap(struct snapshot_rec *rec, __u8 *data, int len);
int snapshot_parse_rmap(struct snapshot_rec *rec, __u8 *data, int len);
int snapshot_parse_smap(struct snapshot_rec *rec, __u8 *data, int len);

int snapshot_rec_alloc_internal(struct snapshot_rec *rec, int maxlinkcnt,
                                int maxstatcnt);
void snapshot_rec_free_internal(struct snapshot_rec *rec);

inline int snapshot_find_link_by_dev_addr(struct snapshot_rec *rec, unsigned char *dev_addr);
inline int snapshot_find_link_by_addr(struct snapshot_rec *rec, __u32 addr);
inline int snapshot_find_stat_by_addr(struct snapshot_rec *rec, __u32 addr);

inline int snapshot_print_state(char *page, struct snapshot_state *s);
inline int snapshot_print_statistic(char *page, struct snapshot_statistic *s);
inline int snapshot_print_rec(char *page, struct snapshot_rec *rec);
void snapshot_clear(struct snapshot_rec *rec);
void snapshot_reset(struct snapshot_rec *rec);

#endif /* _IPT_SNAPSHOT_H */
