/* Shared library add-on to iptables to add EDGEMARK target support. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_EDGEMARK.h>

struct edgemarkinfo {
	struct ipt_entry_target t;
	struct ipt_edgemark_target_info edgemark;
};

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"EDGEMARK target v%s options:\n"
"  --mark-prob value             set the marking probaility (1..99)\n"
"\n",
IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "mark-prob", 1, 0, '1' },
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_edgemark_target_info *edgemarkinfo =
    (struct ipt_edgemark_target_info *)(t)->data;
	*nfcache |= NFC_UNKNOWN;

	/* We assign the average to be 50 which is our default value */
	/* 50 * 2.55 = 128 */
	edgemarkinfo->markprob = 128;
  edgemarkinfo->markprob_unscaled = 50;
}

#define IPT_EDGEMARK_OPT_MARKPROB	0x01

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_edgemark_target_info *edgemarkinfo
		= (struct ipt_edgemark_target_info *)(*target)->data;
  unsigned int num;
  
	switch (c) {
	case '1':
		/* check for common mistakes... */
		if (invert)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify ! --mark-prob");
		if (*flags & IPT_EDGEMARK_OPT_MARKPROB)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --mark-prob twice");

		/* Remember, this function will interpret a leading 0 to be 
		   Octal, a leading 0x to be hexdecimal... */
    if (string_to_number(optarg, 1, 99, &num) == -1 || num < 1)
      exit_error(PARAMETER_PROBLEM,
                 "bad --mark-prob `%s', must be between 1 and 99", optarg);

		/* assign the values */
		edgemarkinfo->markprob = (int)(num * 2.55);
    edgemarkinfo->markprob_unscaled = num;
    *flags |= IPT_EDGEMARK_OPT_MARKPROB;
		break;
	default:
		return 0;
	}
	return 1;
}

static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
		           "EDGEMARK target: Parameter --mark-prob is required");
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	const struct ipt_edgemark_target_info *edgemarkinfo =
		(const struct ipt_edgemark_target_info *)target->data;

//  div_t result = div((edgemarkinfo->markprob*100), 255);
//	if (result.rem > 127)  /* round up... */
//		++result.quot;
//	printf(" marking probability %u%% ", result.quot);
  printf(" marking probability %lu%% ", edgemarkinfo->markprob_unscaled);
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_edgemark_target_info *edgemarkinfo =
		(const struct ipt_edgemark_target_info *)target->data;
//	div_t result = div((edgemarkinfo->markprob*100), 255);
//	if (result.rem > 127)  /* round up... */
//		++result.quot;
//	printf(" marking probability %u%% ", result.quot);
  printf(" marking probability %lu%% ", edgemarkinfo->markprob_unscaled);
}

static
struct iptables_target edgemark
= { NULL,
    "EDGEMARK",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_edgemark_target_info)),
    IPT_ALIGN(sizeof(struct ipt_edgemark_target_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_target(&edgemark);
}
