#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_SNAPSHOT.h>

static struct snapshot_rec ssrec;

static unsigned int
ipt_ssrouter_target(struct sk_buff **pskb,
                    unsigned int hooknum,
                    const struct net_device *in,
                    const struct net_device *out,
                    const void *targinfo,
                    void *userinfo)
{
  struct sk_buff *skb = (*pskb);
  struct iphdr *iph = skb->nh.iph;
  struct snapshot_opt *opt = SNAPSHOT_OPTPTR(iph);
  int iphlen = iph->ihl << 2;
  __u8 *data = ((__u8 *) iph) + iphlen;
  int i, j, flags;
  
  ssrec_lock(ssrec, flags);
  
  switch (snapshot_msgtype(iph)) {
  case SNAPSHOT_MSGTYPE_RESET:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSROUTER", "RESET", in, iph, opt);
      DEBUGP("SSROUTER: reset\n");
      snapshot_reset(&ssrec);
    }
    break;

  case SNAPSHOT_MSGTYPE_CLEAR:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSROUTER", "CLEAR", in, iph, opt);
      if (ssrec.local.state != SNAPSHOT_STATE_UNINITIALIZED) {
        DEBUGP("SSROUTER: clear\n");
        snapshot_clear(&ssrec);
      }
    }
    break;
    
  case SNAPSHOT_MSGTYPE_RMAP:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSROUTER", "RMAP", in, iph, opt);
      DEBUGP("SSROUTER: RMAP size = %d\n", skb->len - iphlen);
      
      if ((ssrec.local.state == SNAPSHOT_STATE_UNINITIALIZED) ||
        (ssrec.local.state == SNAPSHOT_STATE_FINISHED))
      {
        ssrec.local.state = SNAPSHOT_STATE_READY;
        ssrec.local.addr = iph->daddr; // trick: get local addr here
        snapshot_parse_rmap(&ssrec, data, skb->len - iphlen);
      }
      
      snapshot_send_ack(&ssrec, iph->daddr, iph->saddr,
                        SNAPSHOT_MSGTYPE_RMAP_ACK);
    }
    break;
    
  case SNAPSHOT_MSGTYPE_MARKER:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSROUTER", "MARKER", in, iph, opt);

      i = snapshot_find_link_by_dev_addr(&ssrec, (unsigned char *) in->dev_addr);
      if (i != -1) {
        if ((ssrec.local.state == SNAPSHOT_STATE_READY) ||
            (ssrec.local.state == SNAPSHOT_STATE_FINISHED))
        {
          ssrec.local.state = SNAPSHOT_STATE_STARTED;
          ssrec.local.cnt = ssrec.packetcnt;
          ssrec.seqno = ntohl(opt->seqno);
        
          for_each_link(&ssrec, j) {
            ssrec.link[j].state = SNAPSHOT_STATE_STARTED;
            ssrec.link[j].cnt = 0;
            snapshot_send_marker(&ssrec, ssrec.local.addr, ssrec.link[j].addr);
          }

          ssrec.link[i].state = SNAPSHOT_STATE_FINISHED;
          ssrec.finished = 1;
        
        } else {

          if (ssrec.link[i].state == SNAPSHOT_STATE_STARTED) {
            ssrec.link[i].state = SNAPSHOT_STATE_FINISHED;
            ssrec.finished++;
          }
        
        }

        DEBUGP("SSROUTER: channel state = %u\n", ssrec.link[i].cnt);

        if (ssrec.finished == ssrec.linkcnt) {
          ssrec.local.state = SNAPSHOT_STATE_FINISHED;
          snapshot_send_snapshot(&ssrec, ssrec.local.addr, ssrec.victim_addr);
          DEBUGP("SSROUTER: send SNAPSHOT (%u)\n", ssrec.seqno);
        }
      
      } else {
        DEBUGP("SSROUTER: no match link\n");
      }
    }
    break;
    
  case SNAPSHOT_MSGTYPE_SNAPSHOT:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {   
      DEBUGP_SSMSG("SSROUTER", "SNAPSHOT", in, iph, opt);
    }
    break;
    
  default:
    // all snapshot messages, not matter valid or not, are not counted
    if (iph->protocol != IPPROTO_SNAPSHOT) {

      if ((hooknum == NF_IP_POST_ROUTING) && (out != NULL) &&
          ((skb->nfmark & 0xFFFF0000) == 0xFFFF0000))
      {
        ssrec.packetcnt++;
        if (skb->nfmark != 0xFFFFFFFF) {
          i = skb->nfmark & 0xFFFF;
          if (ssrec.link[i].state == SNAPSHOT_STATE_STARTED) {
            ssrec.link[i].cnt++;
          }
        }
        
      } else {
        skb->nfmark = 0xFFFFFFFF;
        
        // FORWARD packets here
        if ((hooknum == NF_IP_FORWARD) && (in != NULL)) {
          DEBUGP("SSROUTER: recv PACKET from %u.%u.%u.%u/", NIPQUAD(iph->saddr));
          DEBUGP_DEV_ADDR((unsigned char *) in->dev_addr, in->addr_len);
          
          i = snapshot_find_link_by_dev_addr(&ssrec, (unsigned char *) in->dev_addr);
          if (i != -1) {
            skb->nfmark = 0xFFFF0000 + i;
          } else {
            DEBUGP("SSROUTER: no match link\n");
          }
        }
      }
      
      /*
      // BOTH OUTPUT and FORWARD packets
      if (((hooknum == NF_IP_FORWARD) || (hooknum == NF_IP_LOCAL_OUT)) && (out != NULL)) {
        ssrec.packetcnt++;
      }

      // FORWARD packets here
      if ((hooknum == NF_IP_FORWARD) && (in != NULL)) {
        DEBUGP("SSROUTER: recv PACKET from %u.%u.%u.%u/", NIPQUAD(iph->saddr));
        DEBUGP_DEV_ADDR((unsigned char *) in->dev_addr, in->addr_len);

        i = snapshot_find_link_by_dev_addr(&ssrec, (unsigned char *) in->dev_addr);
        if (i != -1) {
          if (ssrec.link[i].state == SNAPSHOT_STATE_STARTED) {
            ssrec.link[i].cnt++;
          }
        } else {
          DEBUGP("SSROUTER: no match link\n");
        }
      }
      */
    }
  }

  ssrec_unlock(ssrec, flags);

  return IPT_CONTINUE;
}

static int
ipt_ssrouter_checkentry(const char *tablename,
                        const struct ipt_entry *e,
                        void *targinfo,
                        unsigned int targinfosize,
                        unsigned int hook_mask)
{
  /*
	if (hook_mask & ~((1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_FORWARD) | (1 << NF_IP_LOCAL_OUT) | (1 << NF_IP_POST_ROUTING))) {
		DEBUGP("SSROUTER: bad hook\n");
		return 0;
	}
  */
	if (targinfosize != IPT_ALIGN(0)) {
		DEBUGP("SSROUTER: targinfosize %u != 0\n", targinfosize);
		return 0;
	}

	return 1;
}

static int
ssrouter_proc_read(char *page,
                   char **start,
                   off_t off,
                   int count,
                   int *eof,
                   void *data)
{
  int len, flags;
  
  MOD_INC_USE_COUNT;

  ssrec_lock(ssrec, flags);
  len = snapshot_print_rec(page, &ssrec);
  ssrec_unlock(ssrec, flags);
  
  MOD_DEC_USE_COUNT;

  return len;
}

static struct ipt_target ipt_ssrouter_reg
= { { NULL, NULL }, "SSROUTER", ipt_ssrouter_target, ipt_ssrouter_checkentry,
    NULL, THIS_MODULE };

static const char SSROUTER_PROCFS_PATH[] = "snapshot_router";
extern struct proc_dir_entry *proc_net;

static int __init init(void)
{
  struct proc_dir_entry *proc;
  int ret;

  DEBUGP("SSROUTER: init\n");  
  ret = snapshot_rec_alloc_internal(&ssrec, MAX_LINK_LIMIT, 0);
  if (ret < 0) {
    return ret;
  }
  
  proc = create_proc_entry(SSROUTER_PROCFS_PATH, S_IRUSR | S_IRGRP | S_IROTH, proc_net);
  if (!proc) {
    snapshot_rec_free_internal(&ssrec);
    return -ENOMEM;
  }
  proc->read_proc = ssrouter_proc_read;
  proc->write_proc = NULL;
  proc->owner = THIS_MODULE;
  
	return ipt_register_target(&ipt_ssrouter_reg);
}

static void __exit fini(void)
{
  DEBUGP("SSROUTER: exit\n");
  ipt_unregister_target(&ipt_ssrouter_reg);
  remove_proc_entry(SSROUTER_PROCFS_PATH, proc_net);
  snapshot_rec_free_internal(&ssrec);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
