// indicate this is an ipqext extension
#define _IPQEXT_EXTENSION_

//----------------------------------------------------------------------//
//
//  ipqext extension module auth
//
//----------------------------------------------------------------------//

#include <linux/netfilter.h>
#include <libipq.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <router/ipqext/ipqext.h>
#include <string.h>
#include <lib/util.h>
#include <lib/security/auth.h>

/*
 * define extension name here
 *
 * extension name must match with the output module name
 * ( i.e. ipqext_{extension name} )
 */
const char extension_name[] = "auth";

static const char purpose_name[] = "sslserver";

EVP_MD_CTX ctx;
EVP_PKEY *pkey;

X509_STORE *cert_ctx;
int purpose;
BIO *bio_err;

char param_ca_path[IPQEXT_PARAM_LEN];
char param_ca_file[IPQEXT_PARAM_LEN];
char param_cert_file[IPQEXT_PARAM_LEN];
int param_cert_type;

// signature size = 128 (1024-bit RSA)
#define SIGN_SIZE 128

//----------------------------------------------------------------------//
//
//  module functions
//

static void init_openssl()
{
  ERR_load_crypto_strings();
  OpenSSL_add_all_algorithms();
}

static void fini_openssl()
{
  ERR_free_strings();
  EVP_cleanup();
}


/*
 * ipqext structure
 */
static ipqext_t ext = {
  prev      : NULL,
  next      : NULL,
  handle    : NULL,
  active    : 0,
  name      : extension_name,
  version   : "1.0",
  groupmask : 0,
  init      : &init,
  fini      : &fini,
  proc      : &proc,
  start     : &start,
  stop      : &stop,
  reset     : &reset,
  info      : &info,
  help      : &help,
};

/*
 * packet processing function
 *
 * return: NF_ACCEPT   packet is accepted and leaves ipqext immediately
 *         NF_QUEUE    packet is queued for further ipqext processing
 *         NF_DROP     packet is explicitly dropped
 *
 *         negative verdict to indicate packet is altered
 */
int proc(ipq_packet_msg_t * m)
{
  unsigned int retval = NF_QUEUE;

  struct iphdr *iph = (struct iphdr *) m->payload;
  int iphlen = iph->ihl << 2;
  int payloadlen = ntohs(iph->tot_len) - iphlen - sizeof(struct udphdr);
  int datalen = payloadlen - SIGN_SIZE;

  printf
      ("--------------------------------------------------------------------------------\n");
  printf("payloadlen=%d  datalen=%d\n", payloadlen, datalen);

  if (payloadlen > SIGN_SIZE) {
    unsigned char *sigbuf =
        ((unsigned char *) iph) + iphlen + sizeof(struct udphdr);
    unsigned char *data = sigbuf + SIGN_SIZE;
    int ret;

    ret = auth_verify_buffer(&ctx, pkey, data, datalen, sigbuf, SIGN_SIZE);
    switch (ret) {
    case 1:
      printf("signature is valid\n");
      break;

    case 0:
      printf("signature is invalid\n\n");
      bufdump(sigbuf, SIGN_SIZE);
      bufdump(data, datalen);
      break;

    case -1:
      printf("internal error\n");
      break;
    }

  } else {
    printf("no signature found\n");
  }

  return retval;
}

/*
 * extension start function
 *
 * return on success, -1 or nonzero value on error
 */
int start()
{
  /* verify certificate */
  cert_ctx = auth_cert_verify_init(NULL, param_ca_file);
  purpose = auth_cert_verify_get_purpose_id(purpose_name);
  if (auth_verify_cert(cert_ctx, param_cert_file, purpose, 0) != 1) {
    printf("certificate verification failure\n");
    return -1;
  }

  /* load server public key */
  pkey = auth_load_publickey(param_cert_type, param_cert_file, NULL);
  if (pkey == NULL) {
    if (cert_ctx != NULL)
      X509_STORE_free(cert_ctx);
    printf("failed to retrieve public key from certificate\n");
    return -1;
  }

  return 0;
}

/*
 * extension stop function
 *
 * return on success, -1 or nonzero value on error
 */
int stop()
{
  if (pkey != NULL) {
    EVP_PKEY_free(pkey);
    pkey = NULL;
  }
  if (cert_ctx != NULL) {
    X509_STORE_free(cert_ctx);
    cert_ctx = NULL;
  }
  return 0;
}

/*
 * extension reset function
 *
 * return on success, -1 or nonzero value on error
 */
int reset()
{
  return 0;
}

/*
 * extension info function
 *
 * return a pointer to a char buffer or NULL
 */
const char infobuf[] = " info here ";

const char *info()
{
  return infobuf;
}

/*
 * extension help function
 *
 * return a pointer to a char buffer or NULL
 */
const char helpbuf[] = " help here ";

const char *help()
{
  DB("%s: help invoked\n", extension_name);
  return helpbuf;
}

/*
 *  initial pass function
 *
 * return on success, -1 or nonzero value on error
 */
int init_pass(void *handle, char *param, int paramlen)
{
  char temp[IPQEXT_PARAM_LEN];
  int i, st, ed, idx;

  // set module handle
  ext.handle = handle;

  // parse parameter buffer
  if ((param == NULL) || (paramlen == 0))
    return -1;

  st = 0;
  idx = 0;

  for (i = 0; i < paramlen; i++) {
    if ((param[i] == ',') || (param[i] == '\0')) {
      ed = i;

      if (ed - st > 0) {

        switch (idx) {
        case 0:
          strncpy(param_ca_path, &(param[st]), ed - st);
          param_ca_path[ed - st] = '\0';
          idx++;
          break;

        case 1:
          strncpy(param_ca_file, &(param[st]), ed - st);
          param_ca_file[ed - st] = '\0';
          idx++;
          break;

        case 2:
          strncpy(param_cert_file, &(param[st]), ed - st);
          param_cert_file[ed - st] = '\0';
          idx++;
          break;

        case 3:
          strncpy(temp, &(param[st]), ed - st);
          temp[ed - st] = '\0';
          param_cert_type = atoi(temp);
          if ((param_cert_type < 1) || (param_cert_type > 3)) {
            CERR("invalid certificate type\n");
            return -1;
          }
          return 0;
          break;
        }
      }
      st = i + 1;
    }

    if (param[i] == '\0')
      break;
  }

  return -1;
}

/*
 * extension initalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int init()
{
  // initalization codes here

  /* OpenSSL initialization */
  init_openssl();
  auth_init_bio_err();

  // register function must always be the last call
  ipqext_register(&ext);

  return 0;
}

/*
 * extension finalization function
 *
 * return 0 on success, -1 or nonzero value on error
 */
int fini()
{
  // unregister function must always be the first call
  ipqext_unregister(&ext);

  // finialization codes here

  /* OpenSSL finalization */
  auth_free_bio_err();
  fini_openssl();

  return 0;
}

/*
 * module initalization function
 */
void _init(void)
{
}

/*
 * module finalization function
 */
void _fini(void)
{
}

//----------------------------------------------------------------------//
