//----------------------------------------------------------------------//
//
//  REMARKS
//
//  dq_buffer_t.size == dq_block.b_size
//  add enough lockings
//  parametets checking
//
//----------------------------------------------------------------------//

#define __DEBUG__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/time.h>
#include <router/ipqext/extensions/ipqext_fec/fec.h>
#include <router/ipqext/extensions/ipqext_fec/decqueue.h>

static inline void __dq_add_block(dq_block_t *prev, dq_block_t *blk);
static inline void __dq_del_block(dq_block_t *blk);

#define foreach_block_in_queue(que, blk) \
  for (blk = que->head.next; blk != &(que->head); blk = blk->next)


//----------------------------------------------------------------------//
//
//  dq_buffer
//

int dq_new_copy_buffer(dq_buffer_t **bufptr, unsigned char *data,
                       unsigned int size)
{
  dq_buffer_t *buf = *bufptr;

  buf = (dq_buffer_t*) malloc(sizeof(dq_buffer_t));
  if (buf == NULL) return -1;

  buf->buf = (unsigned char*) malloc(size * sizeof(unsigned char));
  if (buf->buf == NULL) {
    free(buf);
    return -1;
  }

  if (data != NULL) {
    memcpy(buf->buf, data, size);
  }

  buf->size = size;
  *bufptr = buf;

  return 0;
}

int dq_new_buffer_link(dq_buffer_t **bufptr, unsigned char *link,
                       unsigned int size)
{
  dq_buffer_t *buf = *bufptr;

  buf = (dq_buffer_t*) malloc(sizeof(dq_buffer_t));
  if (buf == NULL) return -1;

  buf->buf = link;
  buf->size = size;
  *bufptr = buf;

  return 0;
}

void dq_free_buffer(dq_buffer_t **bufptr)
{
  dq_buffer_t *buf = *bufptr;

  if (buf == NULL)
    return;

  if (buf->buf != NULL)
    free(buf->buf);

  free(buf);
  *bufptr = NULL;
}


//----------------------------------------------------------------------//
//
//  dq_block
//

int dq_new_block(dq_block_t **blkptr, unsigned int id, unsigned int k_size,
                 unsigned int n_size, unsigned int b_size)
{
  dq_block_t *blk;

  blk = (dq_block_t*) malloc(sizeof(dq_block_t));
  if (blk == NULL) return -1;

  if (pthread_mutex_init(&(blk->mutex), NULL) != 0) {
    free(blk);
    return -1;
  }

  if (pthread_mutex_init(&(blk->nbuffer_mutex), NULL) != 0) {
    pthread_mutex_destroy(&(blk->mutex));
    free(blk);
    return -1;
  }

  if (pthread_cond_init(&(blk->nbuffer_cond), NULL) != 0) {
    pthread_mutex_destroy(&(blk->mutex));
    pthread_mutex_destroy(&(blk->nbuffer_mutex));
    free(blk);
    return -1;
  }

  blk->buflist = (dq_buffer_t**) malloc(n_size * sizeof(dq_buffer_t*));
  if (blk->buflist == NULL) {
    pthread_mutex_destroy(&(blk->mutex));
    pthread_mutex_destroy(&(blk->nbuffer_mutex));
    pthread_cond_destroy(&(blk->nbuffer_cond));
    free(blk);
    return -1;
  }
  memset(blk->buflist, 0, n_size * sizeof(dq_buffer_t*));

  blk->id = id;
  blk->k_size = k_size;
  blk->n_size = n_size;
  blk->b_size = b_size;
  blk->nbuffer = 0;
  blk->prev = NULL;
  blk->next = NULL;
  blk->que = NULL;
  gettimeofday(&(blk->timestamp), NULL);
  *blkptr = blk;

  return 0;
}

void dq_free_block(dq_block_t **blkptr)
{
  dq_block_t *blk = *blkptr;
  int i;

  if (blk == NULL)
    return;

  if (blk->buflist != NULL) {
    for (i=0; i<blk->n_size; i++) {
      dq_free_buffer(&(blk->buflist[i]));
    }
    free(blk->buflist);
  }

  pthread_mutex_destroy(&(blk->mutex));
  pthread_mutex_destroy(&(blk->nbuffer_mutex));
  pthread_cond_destroy(&(blk->nbuffer_cond));
  free(blk);
  *blkptr = NULL;
}

void dq_show_block(dq_block_t *blk)
{
  int i;

  if (blk == NULL) return;

  dq_lock_block(blk);

  printf("[");
  printf("id:%u  k_size:%u  n_size:%u  b_size:%u  nbuffer:%d  buflist:",
         blk->id, blk->k_size, blk->n_size, blk->b_size, blk->nbuffer);
  if (blk->nbuffer == 0) {
    printf("NULL");
  } else {
    for (i=0; i<blk->n_size; i++)
      printf("%c", (blk->buflist[i] == NULL) ? '-' : '#');
  }
  printf("]\n");

  dq_unlock_block(blk);
}

int dq_decode_block(dq_block_t *blk, dq_decinfo_t *dec, void *fec)
{
  int i, j, cnt, contiguous;
  int ret = 0;

  /*
    ret
    0 : success
    -1: error
    1 : incomplete block
  */

  if ((blk == NULL) || (dec == NULL) || (dec->buf == NULL) ||
      (dec->index == NULL) || (blk->nbuffer == 0)) {
    return -1;
  }

  j = 0;
  cnt = 0;
  contiguous = 0;

  for (i=0; (i<blk->n_size) && (j<blk->k_size) && (!contiguous); i++) {
    if (blk->buflist[i] != NULL) {
      dec->buf[j] = blk->buflist[i]->buf;
      dec->index[j] = i;
      j++;
      cnt++;
    }
    if ((i == blk->k_size-1) && (cnt == blk->k_size))
      contiguous = 1;
  }

  if (cnt < blk->k_size) {
    printf("[incomplete block]   ");
    dq_show_block(blk);
    return 1;   // incomplete block
  }

  if (!contiguous) {
    // decode
    //    printf("FEC decode block\n");
    ret = fec_decode(fec, (void *) dec->buf,
                     dec->index, blk->b_size);
  } else {
    //    printf("contiguous block\n");
  }

  // exit:
  return ret;
}

inline int dq_istimetoleave_block(dq_block_t *blk)
{
  /*
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (blk->nbuffer >= blk->k_size) ||
    (((tv.tv_sec - blk->timestamp.tv_sec) * 1000000 +
      (tv.tv_usec - blk->timestamp.tv_usec)) >= DQ_BLOCK_EXPIRYTIME);
  */

  long deta_sec, deta_usec;

  struct timeval tv;
  struct timespec ts;

  dq_lock_block_nbuffer(blk);

  gettimeofday(&tv, NULL);
  deta_sec = tv.tv_sec - blk->timestamp.tv_sec;
  deta_usec = tv.tv_usec - blk->timestamp.tv_usec;

  while ((blk->nbuffer < blk->k_size) &&
         ((deta_sec * 1000000 + deta_usec) < DQ_BLOCK_EXPIRYTIME)) {
    
    ts.tv_sec = 0;
    ts.tv_nsec = (DQ_BLOCK_EXPIRYTIME - deta_usec) * 1000;
    ts.tv_nsec = (ts.tv_nsec < 20000000) ? 20000000 : ts.tv_nsec;
    pthread_cond_timedwait(&(blk->nbuffer_cond), &(blk->nbuffer_mutex), &ts);
    //    pthread_cond_wait(&(blk->nbuffer_cond), &(blk->nbuffer_mutex));
    
    gettimeofday(&tv, NULL);
    deta_sec = tv.tv_sec - blk->timestamp.tv_sec;
    deta_usec = tv.tv_usec - blk->timestamp.tv_usec;
  }

  dq_unlock_block_nbuffer(blk);

  return 1;
}

static inline void __dq_add_block(dq_block_t *prev, dq_block_t *blk)
{
  /*
    prev - prev->next
    prev - blk - prev->next
  */
  blk->prev = prev;
  blk->next = prev->next;
  blk->prev->next = blk;
  blk->next->prev = blk;
}

static inline void __dq_del_block(dq_block_t *blk)
{
  /*
    blk->prev - blk - blk->next
    blk->prev - blk->next
   */
  blk->prev->next = blk->next;
  blk->next->prev = blk->prev;
  blk->prev = blk;
  blk->next = blk;
}


//----------------------------------------------------------------------//
//
//  dq_decinfo
//

int dq_new_decinfo(dq_decinfo_t **decptr, unsigned int size)
{
  dq_decinfo_t *dec;

  dec = (dq_decinfo_t*) malloc(sizeof(dq_decinfo_t));
  if (dec == NULL) return -1;

  dec->buf = (unsigned char **) malloc(size * sizeof(unsigned char *));
  if (dec->buf == NULL) {
    free(dec);
    return -1;
  }
  memset(dec->buf, 0, size * sizeof(unsigned char *));

  dec->index = (unsigned int *) malloc(size * sizeof(unsigned int));
  if (dec->index == NULL) {
    free(dec->buf);
    free(dec);
    return -1;
  }

  *decptr = dec;

  return 0;
}

void dq_free_decinfo(dq_decinfo_t **decptr)
{
  dq_decinfo_t *dec = *decptr;

  if (dec == NULL)
    return;

  if (dec->buf != NULL)
    free(dec->buf);

  if (dec->index != NULL)
    free(dec->index);

  free(dec);
  *decptr = NULL;
}


//----------------------------------------------------------------------//
//
//  dq_queue
//

int dq_new_queue(dq_queue_t **queptr, unsigned int headid, unsigned int wsize,
                 unsigned int k_size, unsigned int n_size,
                 __u32 src_addr, __u32 dst_addr)
{
  dq_queue_t *que = *queptr;

  que = (dq_queue_t*) malloc(sizeof(dq_queue_t));
  if (que == NULL) return -1;

  if (pthread_mutex_init(&(que->mutex), NULL) != 0) {
    free(que);
    return -1;
  }

  if (pthread_mutex_init(&(que->nblock_mutex), NULL) != 0) {
    pthread_mutex_destroy(&(que->mutex));
    free(que);
    return -1;
  }

  if (pthread_cond_init(&(que->nblock_cond), NULL) != 0) {
    pthread_mutex_destroy(&(que->nblock_mutex));
    pthread_mutex_destroy(&(que->mutex));
    free(que);
    return -1;
  }

  // initialize FEC
  que->fec = (void *) fec_new(k_size, n_size);
  if (que->fec == NULL) {
    pthread_cond_destroy(&(que->nblock_cond));
    pthread_mutex_destroy(&(que->nblock_mutex));
    pthread_mutex_destroy(&(que->mutex));
    free(que);
    return -1;
  }

  que->k_size = k_size;
  que->n_size = n_size;

  que->nblock = 0;
  que->headid = headid;
  que->wsize = wsize;
  que->head.prev = &(que->head);
  que->head.next = &(que->head);
  que->src_addr = src_addr;
  que->dst_addr = dst_addr;
  *queptr = que;

  return 0;
}

void dq_free_queue(dq_queue_t **queptr)
{
  dq_queue_t *que = *queptr;
  dq_block_t *blk, *next;

  for (blk = que->head.next; blk != &(que->head); blk = next) {
    next = blk->next;
    dq_free_block(&blk);
  }
  fec_free(que->fec);
  pthread_cond_destroy(&(que->nblock_cond));
  pthread_mutex_destroy(&(que->nblock_mutex));
  pthread_mutex_destroy(&(que->mutex));
  free(que);
  *queptr = NULL;
}

void dq_show_queue(dq_queue_t *que)
{
  dq_block_t *blk;

  if (que == NULL) return;

  dq_lock_queue(que);

  printf("----------------------------------------------------------------------\n");
  printf("nblock:%d  headid:%u  wsize:%u  ksize:%u  nsize:%u\n",
         que->nblock, que->headid, que->wsize, que->k_size, que->n_size);
  if (dq_isempty_queue(que)) {
    printf("EMPTY QUEUE\n");
  } else {
    foreach_block_in_queue(que, blk) {
      dq_show_block(blk);
    }
  }
  printf("----------------------------------------------------------------------\n");

  dq_unlock_queue(que);
}

void dq_show_queue_status(dq_queue_t *que)
{
  dq_block_t *blk;

  if (que == NULL) return;

  dq_lock_queue(que);

  printf("nblock:%d  headid:%u  wsize:%u  ksize:%u  nsize%u  blklist:",
         que->nblock, que->headid, que->wsize, que->k_size, que->n_size);
  if (dq_isempty_queue(que)) {
    printf("EMPTY QUEUE\n");
  } else {
    foreach_block_in_queue(que, blk) {
      dq_lock_block(blk);
      if (blk->nbuffer == blk->k_size)
        printf("#");
      else if (blk->nbuffer < blk->k_size)
        printf("*");
      else
        printf("-");
      dq_unlock_block(blk);
    }
  }
  printf("\n");
  dq_unlock_queue(que);
}

inline int dq_isaccept_blockid(dq_queue_t *que, unsigned int id)
{
  /*
    -----##########+------
    0    H         M     F

    ###+-----------#######
    0  M           H     F
   */
  int ret;
  unsigned int max = que->headid + que->wsize;

  if (max < que->headid) {
    ret = (id >= que->headid) || (id < max);
  } else {
    ret = (id >= que->headid) && (id < max);
  }
  return ret;
}


//----------------------------------------------------------------------//
//
//  dq_buffer-to-dq_block functions
//

int dq_add_buffer(dq_block_t *blk, int idx, dq_buffer_t *buf)
{
  if (blk->buflist[idx] != NULL)
    return -1;

  if ((idx < 0) || (idx >= blk->n_size))
    return -1;

  dq_lock_block_nbuffer(blk);
  blk->nbuffer++;
  pthread_cond_signal(&(blk->nbuffer_cond));
  dq_unlock_block_nbuffer(blk);

  blk->buflist[idx] = buf;

  return 0;
}

dq_buffer_t *dq_del_buffer(dq_block_t *blk, int idx)
{
  dq_buffer_t *buf;

  if (blk->buflist[idx] == NULL)
    return NULL;

  if ((idx < 0) || (idx >= blk->n_size))
    return NULL;
  
  buf = blk->buflist[idx];
  blk->buflist[idx] = NULL;

  dq_lock_block_nbuffer(blk);
  blk->nbuffer--;
  dq_unlock_block_nbuffer(blk);

  return buf;
}


//----------------------------------------------------------------------//
//
//  dq_block-to-dq_queue functions
//

dq_block_t *dq_get_block(dq_queue_t *que, unsigned int id)
{
  dq_block_t *blk;

  dq_lock_queue(que);

  foreach_block_in_queue(que, blk) {
    if (blk->id == id)
      break;
  }

  if (blk == &(que->head))
    blk = NULL;

  dq_unlock_queue(que);

  return blk;
}

int dq_enqueue_block(dq_queue_t *que, dq_block_t *blk)
{
  dq_block_t *tmp;
  
  dq_lock_queue(que);

  if (!dq_isaccept_blockid(que, blk->id)) {
    dq_unlock_queue(que);
    return -1;
  }

  foreach_block_in_queue(que, tmp) {
    if (blk->id == tmp->id) {
      dq_unlock_queue(que);
      return -1;
    }

    if (blk->id < tmp->id)
      break;
  }

  __dq_add_block(tmp->prev, blk);

  // signal condition variable (nblock)
  dq_lock_queue_nblock(que);
  que->nblock++;
  pthread_cond_signal(&(que->nblock_cond));
  dq_unlock_queue_nblock(que);

  dq_lock_block(blk);
  blk->que = que;
  dq_unlock_block(blk);

  dq_unlock_queue(que);

  return 0;
}

int dq_enqueue_create_block(dq_queue_t *que, dq_block_t **blkptr,
                            unsigned int id, unsigned int b_size)
{
  dq_block_t *tmp, *blk;
  
  dq_lock_queue(que);

  if (!dq_isaccept_blockid(que, id)) {
    dq_unlock_queue(que);
    return -1;
  }

  if (dq_new_block(&blk, id, que->k_size, que->n_size, b_size) != 0) {
    dq_unlock_queue(que);
    return -1;
  }

  foreach_block_in_queue(que, tmp) {
    if (blk->id == tmp->id) {
      dq_unlock_queue(que);
      return -1;
    }

    if (blk->id < tmp->id)
      break;
  }
  __dq_add_block(tmp->prev, blk);


  // signal condition variable (nblock)
  dq_lock_queue_nblock(que);
  que->nblock++;
  pthread_cond_signal(&(que->nblock_cond));
  dq_unlock_queue_nblock(que);

  dq_lock_block(blk);
  blk->que = que;
  dq_unlock_block(blk);

  *blkptr = blk;

  dq_unlock_queue(que);

  return 0;
}

dq_block_t *dq_dequeue_block(dq_queue_t *que)
{
  /*
  dq_block_t *blk = NULL;

  // wait until there is a block
  dq_lock_queue_nblock(que);  
  while (que->nblock == 0) {
    pthread_cond_wait(&(que->nblock_cond), &(que->nblock_mutex));
  }
  dq_unlock_queue_nblock(que);

  dq_lock_queue(que);
  blk = que->head.next;
  dq_unlock_queue(que);
  
  dq_istimetoleave_block(blk);

  dq_lock_queue(que);

  __dq_del_block(blk);
  
  dq_lock_block(blk);
  blk->que = NULL;
  dq_unlock_block(blk);
  
  dq_lock_queue_nblock(que);
  que->nblock--;
  dq_unlock_queue_nblock(que);
  
  // advance head
  que->headid = (que->head.next != NULL) ?
    (que->headid+1) : que->head.next->id;
  
  dq_unlock_queue(que);
  */
  /*
  dq_lock_queue(que);

  blk = que->head.next;
  if (dq_istimetoleave_block(blk)) {
    __dq_del_block(que->head.next);
    
    dq_lock_block(blk);
    blk->que = NULL;
    dq_unlock_block(blk);
    
    dq_lock_queue_nblock(que);
    que->nblock--;
    dq_unlock_queue_nblock(que);
    
    // advance head
    que->headid = (que->head.next != NULL) ?
      (que->headid+1) : que->head.next->id;
  } else {
    blk = NULL;
  }

  dq_unlock_queue(que);
  */

  dq_block_t *blk = NULL;

  dq_lock_queue_nblock(que);
  while (que->nblock <= 0) {
    pthread_cond_wait(&(que->nblock_cond), &(que->nblock_mutex));
  }
  dq_unlock_queue_nblock(que);

  dq_lock_queue(que);
  blk = que->head.next;
  
  dq_unlock_queue(que);
  dq_istimetoleave_block(blk);
  dq_lock_queue(que);

  __dq_del_block(que->head.next);

  dq_lock_block(blk);
  blk->que = NULL;
  dq_unlock_block(blk);
  
  que->nblock--;
    
  // advance head
  que->headid = (que->head.next != NULL) ?
      (que->headid+1) : que->head.next->id;

  //  printf("%d\n", que->nblock);
  
  dq_unlock_queue(que);

  return blk;
}
